/* 

IBM Confidential 
	
(c) Copyright IBM Corp. 2000, 2008 All Rights Reserved.

The source code for this program is not published or otherwise 
divested of its trade secrets, irrespective of what has been 
deposited with the U.S. Copyright Office.

*/ 

/*
 * Licensed Materials - Use restricted. For use in compiling interfaces to
 * IBM Rational TestManager.
 */

/*
	.../include/sme/data.h			(formatted 4 spaces per tab)
	
	This header file contains procedures for saving and
	parsing the row data returned during an sqlnrecv or
	sqlfetch_cursor command.

	The only behavior modifier defined is _PV_DATA_REBUILD which,
	when defined, causes SaveData to compute field parameters even 
	if the specified data set already exists.
*/

#ifndef	_PV_DATA_H		/* prevent duplicate inclusion */
#define	_PV_DATA_H

/*
	Remove leading and trailing blanks from a text string.
	If the string isn't null to start with, a single blank will
	be returned for completely blank values.
*/
string func trim_data(smes_trm_text)
string smes_trm_text;
{
	if (smes_trm_text == "") return ("");

	for (smen_trm_index = strlen(smes_trm_text); 
		 smen_trm_index > 0;
		 smen_trm_index--)
		{
		if (stoc(smes_trm_text,smen_trm_index) != ' ')
			{
			break;
			}
		}
	for (smen_trm_start = 1; smen_trm_start < smen_trm_index; smen_trm_start++)
		{
		if (stoc(smes_trm_text,smen_trm_start) != ' ')
			{
			break;
			}
		}

	if (smen_trm_start > smen_trm_index)
		{
		return (" ");
		}

	return (substr(smes_trm_text,smen_trm_start,
								 smen_trm_index - smen_trm_start + 1));
}

/*
	Various data structures needed to store the returned rows
	and field sizes.  These routines assume that _column_headers
	consists of a \n terminated line of field headers, followed
	by a \n terminated line of ===== characters.  This line
	is assumed to have a single space between each field heading,
	and fixed length fields that can be sized by counting the
	number of "=" characters.  Data rows are assumed to be a set
	of fixed length records, with a terminating \n.

	The various data structures are:

	smes_dat_names  array of saved data names.
	smes_dat_data   array of saved data text.
	smen_dat_start  array (dual index) of starting columns for fields
	smen_dat_length array (dual index) of field lengths

	In smen_dat_length, the 0 index is the total record length.
	In smen_dat_start, the 0 index is the number of fields.
*/
#define SMEN_DAT_FIELDS(X)  smen_dat_start[X][0]
#define SMEN_DAT_RECLENGTH(X)   smen_dat_length[X][0]

persistent string smes_dat_names[] = {""};
persistent string smes_dat_data[];
persistent int smen_dat_start[][];
persistent int smen_dat_length[][];

/*
	This routine finds the index associated with a particular
	data name.  If one is not found, and the value of smen_ddi_create
	is non-zero, a new one is allocated.  smes_dat_data and
	smen_dat_start are set to nulls.
*/
func DataIndex(smes_ddi_name, smen_ddi_create)
string smes_ddi_name;
{
	/* smen_ddi_index is where we'll add entry if needed */
	smen_ddi_index = -1;

	/* search for an existing entry */

	for (smen_dat_index = 0;
		 smen_dat_index <= limitof smes_dat_names;
		 smen_dat_index++)
		{
		if (smes_ddi_name == smes_dat_names[smen_dat_index])
			{
			return (smen_dat_index);
			}

		/* remember first empty slot for recycling */
		if ((smen_ddi_index == -1) &&
			(smes_dat_names[smen_dat_index] == ""))
			{
			smen_ddi_index = smen_dat_index;
			}
		}

	/* Not found, return -1 if create not invoked. */
	if (!smen_ddi_create) return (-1);
	
	/* add a slot if no empty ones were found */
	if (smen_ddi_index == -1)
		{
		smen_ddi_index = (limitof smes_dat_names)+1;
		}

	/* initialize data structures */
	smes_dat_names[smen_ddi_index] = smes_ddi_name;
	SMEN_DAT_FIELDS(smen_ddi_index) = 0;

	return (smen_ddi_index);
}

/*
	Free (as much as possible) the space associated with a set of data.  
	smes_dfd_name tells which string to free.
*/
proc FreeData(smes_dfd_name)
string smes_dfd_name;
{
	/* get the index, just return if non-existent */
	if ((smen_dat_index = DataIndex(smes_dfd_name,0)) == -1)
		{
		return;
		}

	/* clear the data structures */
	smes_dat_names[smen_dat_index] = "";
	smes_dat_data[smen_dat_index] = "";
	SMEN_DAT_FIELDS(smen_dat_index) = 0;

	return;
}

/*
	This routine saves the data in _response, and
	sets the data parameters according to the field
	delimiters found in _column_headers  For efficiency purposes,
	it does not rescan the fields if a valid entry exists
	using that name, unless _PV_DATA_REBUILD is defined.
*/
proc SaveData(smes_dsd_name)
string smes_dsd_name;
{
	string smes_dsd_spacer, smes_dsd_spacers;

	/* get the index */
	smen_dat_index = DataIndex(smes_dsd_name,1);
	
	/* save the data */
	smes_dat_data[smen_dat_index] = _response;

#ifndef	_PV_DATA_REBUILD
	if (SMEN_DAT_FIELDS(smen_dat_index) != 0)
		{
		return;
		}
#endif

	/* initialize the record length */
	SMEN_DAT_RECLENGTH(smen_dat_index) = 1;

	/* get the heading/data separator */
	smes_dsd_spacers = subfield(_column_headers,"\n",2);

	/* step through the fields, saving the data */
	for (smen_dsd_index = 1;
		 (smes_dsd_spacer=subfield(smes_dsd_spacers," ",smen_dsd_index)) != "";
		 smen_dsd_index++)
		{
		smen_dat_start[smen_dat_index][smen_dsd_index] =
						SMEN_DAT_RECLENGTH(smen_dat_index);
		smen_dat_length[smen_dat_index][smen_dsd_index] =
						strlen(smes_dsd_spacer);
		SMEN_DAT_RECLENGTH(smen_dat_index) += strlen(smes_dsd_spacer) + 1;
		}

	if (substr(_response,SMEN_DAT_RECLENGTH(smen_dat_index),1) != "\n")
		{
		smen_dsd_temp = SMEN_DAT_RECLENGTH(smen_dat_index) - 1;
		SMEN_DAT_RECLENGTH(smen_dat_index) = smen_dsd_temp;
		}

	SMEN_DAT_FIELDS(smen_dat_index) = smen_dsd_index - 1;

	return;
}

/*
	This routine appends the data in _response to a previously
	set of data saved by SaveData.  If no such data exists, it
	calls the SaveData routine to create one.  it returns the
	total length of the saved data, including the appended data.
*/
func AppendData(smes_dad_name)
string smes_dad_name;
{
	/* get the index, calling SaveData if invalid */
	if ((smen_dat_index = DataIndex(smes_dad_name,0)) == -1)
		{
		SaveData(smes_dad_name);
		smen_dat_index = DataIndex(smes_dad_name,0);
		}
	else
		{						/* append the data */
		smes_dat_data[smen_dat_index] += _response;
		}
	return (strlen(smes_dat_data[smen_dat_index]));
}

/*
	This function returns the specified record and field from
	the requested data block.
*/
string func GetData(smes_dgf_name, smen_dgf_record, smen_dgf_field)
string smes_dgf_name;
{
	/* get the index, return null if non-existent */
	if ((smen_dat_index = DataIndex(smes_dgf_name,0)) == -1)
		{
		return ("");
		}

	/* check to see if the field number is valid */
	if (smen_dgf_field > SMEN_DAT_FIELDS(smen_dat_index))
		{
		return ("");
		}

	/* retrieve and return the data field */
	return(trim_data(substr(smes_dat_data[smen_dat_index],
					   (smen_dgf_record-1)*SMEN_DAT_RECLENGTH(smen_dat_index)+
					   smen_dat_start[smen_dat_index][smen_dgf_field],
					   smen_dat_length[smen_dat_index][smen_dgf_field])));
}

/*
	This func returns the specified field when there is only
	one record in the data.
*/
string func GetData1(smes_dg1_name, smen_dg1_field)
string smes_dg1_name;
{
	/* get the index, return null if not valid */
	if ((smen_dat_index = DataIndex(smes_dg1_name, 0)) == -1)
		{
		return ("");
		}

	/* check to see if the field number is valid */
	if (smen_dg1_field > SMEN_DAT_FIELDS(smen_dat_index))
		{
		return ("");
		}

	/* retrieve and return the data field */
	return(trim_data(substr(smes_dat_data[smen_dat_index],
					   smen_dat_start[smen_dat_index][smen_dg1_field],
					   smen_dat_length[smen_dat_index][smen_dg1_field])));
}

/*
	Clear all data save areas.
*/
proc FreeAllData()
{
	for (smen_fad_index = 0;
		 smen_fad_index <= limitof smes_dat_names;
		 smen_fad_index++)
		{
		smes_dat_names[smen_fad_index] = "";
		smes_dat_data[smen_fad_index] = "";
		SMEN_DAT_FIELDS(smen_fad_index) = 0;
		}
	return;
}

#endif			/* matches #ifndef _PV_DATA_H at top */
