/*
 *
 *	Copyright (C) Rational Software Corporation, 1993, 1994, 1995, 1996,
 *	1997, 1998, 1999
 *	ALL RIGHTS RESERVED WORLDWIDE.
 *
 *	This Module contains Proprietary Information of
 *	Rational Software Corporation
 *	and should be treated as Confidential.
 *
 *	VU is a trademark of Rational Software Corporation.
 *
 */

/*
	.../include/sme/fileio.h		(formatted 4 spaces per tab)
	
	This header includes routines for standardized reading of files of
	ASCII data, where the files are organized by lines.  It also has
	routines for splitting the lines into fields and subfields.

	Note that in most cases, the functionality of these routines can be
	achieved with higher efficiency and flexibility in the VU language
	built-in datapool routines.
*/

#ifndef	_PV_FILEIO		/* guard against recursive includes */
#define	_PV_FILEIO

/*
	If we have VU version 4.0 or later, and the user has defined the
	macro symbol _PV_FILEIO_ALLOWNULLS, then we will provide an
	alternative function to the VU subfield function that will
	treat consecutive delimiters as indicating a null subfield.
	This will affect the operation of all fileio.h functions that
	use the VU subfield function, when the value of Allow_Nulls
	is non-zero.
*/
#ifdef	_VU_VERSION				/* check for VU 4.0 or later */
#ifdef	_PV_FILEIO_ALLOWNULLS	/* user wants null processing ability */

persistent int Allow_Nulls = 0;		/* flag to look for nulls */
persistent string smes_sf_lasttext = "";	/* last text that was parsed */
persistent string smes_sf_lastdelim = "";	/* last set of delimiters */
persistent int smen_sf_fields = 0;		/* fields in last string */
persistent string smes_sf_value[];		/* actual values parsed */
persistent int Final_Field = 0;			/* flag for last field */
persistent int Final_SubField = 0;		/* flag set for last subfield */

/*
	This function is similar to the built-in VU subfield function,
	except that it considers two adjacent delimiters as a null string.
	For efficiency purposes, the entire string is passed during the
	call, and if the next call has the same text and delimiters, no
	parsing is performed.
*/
string func SubField(smes_sf_text, smes_sf_delimiters, smen_sf_field)
string smes_sf_text, smes_sf_delimiters;
{
/*
	Parse only if using new values.
*/
if ((smes_sf_text != smes_sf_lasttext) ||
	(smes_sf_delimiters != smes_sf_lastdelim))
	{
	smes_sf_lasttext = smes_sf_text;
	smes_sf_lastdelim = smes_sf_delimiters;
	smen_sf_start = 1;			/* first character of current subfield */
	smen_sf_length = strlen(smes_sf_text);	/* only compute once */
	smen_sf_fields = 0;			/* none parsed yet. */
	/*
		Check each chearacter to see if it's a delimiter.
	*/
	for (smen_sf_index = 1; smen_sf_index <= smen_sf_length; smen_sf_index++)
		{
		if (cindex(smes_sf_delimiters,stoc(smes_sf_text,smen_sf_index)) != 0)
			{
			/* cut the string out and save it */
			smes_sf_value[smen_sf_fields++] =
				substr(smes_sf_text,smen_sf_start,smen_sf_index-smen_sf_start);
			/* mark start of next subfield */
			smen_sf_start = smen_sf_index + 1;
			}
		}
	/*
		Save the last field.
	*/
	smes_sf_value[smen_sf_fields++] =
		substr(smes_sf_text,smen_sf_start,smen_sf_index-smen_sf_start);
	}
/*
	Return the specified subfield, or a null if the index is invalid.
*/
if ((smen_sf_field > smen_sf_fields) || (smen_sf_field < 0))
	{
	return ("");
	}
return(smes_sf_value[smen_sf_field-1]);
}

#endif			/* matches #ifdef _PV_FILEIO_ALLOWNULLS above. */
#endif			/* matches #ifdef _VU_VERSION above. */

/*
	The function ReadLine reads the next line of input from the passed
	file descriptor.  As part of the read function, comment and blank
	lines are ignored, and the read will wrap around the end of the file
	if _PV_FILEIO_NOWRAP is not defined.

	The function returns a 1 if the read was successful, or a -1 if it
	failed.  After the read, Last_Line contains the value read.

	The variable _PV_FILEIO_COMMENT defines the list of characters which
	are considered to be comment delimiters.  _PV_FILEIO_FIELD and
	_PV_FILEIO_SUBFIELD define the field and subfield delimiting
	characters respectively.  The string given in _PV_FILEIO_WHITESPACE lists
	all "white space" characters which should be removed from the end of
	lines before they are returned.  The defaults for these variables are
	defined below.
*/
#ifndef	_PV_FILEIO_COMMENT
#define	_PV_FILEIO_COMMENT	"#"
#endif

#ifndef	_PV_FILEIO_FIELD
#define _PV_FILEIO_FIELD	"|"
#endif

#ifndef	_PV_FILEIO_SUBFIELD
#define	_PV_FILEIO_SUBFIELD	":"
#endif

#ifndef	_PV_FILEIO_WHITESPACE
#define	_PV_FILEIO_WHITESPACE	" \t"
#endif

/*
	_PV_FILEIO_BLANKLINE is a special string that indicates that an
	empty line should be returned.  Note that setting it to "" will
	cause all empty (and comment) lines to be returned as blank lines.
	The deafult setting is defined below.
*/
#ifndef	_PV_FILEIO_BLANKLINE
#define	_PV_FILEIO_BLANKLINE	"~"
#endif

#ifdef	_VU_VERSION			/* under 4.0 or later, make persistent */
persistent string Last_Line;
persistent string Last_Field;
persistent string Last_SubField;
persistent int sme_next_field;
persistent int sme_next_subfield;
#else
string Last_Line;
string Last_Field;
string Last_SubField;
#endif

func ReadLine(smen_fd)
{
string smes_rl_temp;

Last_Line = "";
while (Last_Line == "")		/* keep trying until we get it */
	{
	smen_offset = ftell(smen_fd);	/* remember where we are */
	smes_rl_temp = "";
	smen_response = fscanf(smen_fd,"%[^\n]",&smes_rl_temp);
	if (smen_response == 1)			/* something read, check length */
		{
		while ((smen_response == 1) && (strlen(smes_rl_temp) == 100))
			{
			Last_Line += smes_rl_temp;
			smen_response = fscanf(smen_fd,"%[^\n]",&smes_rl_temp);
			}
		Last_Line += smes_rl_temp;
		}
	else if (smen_response == 0)	/* nothing read, check pointer */
		{
		if (ftell(smen_fd) == smen_offset)	/* didn't move, unstick it */
			{
			fseek(smen_fd,1,1);
			Last_Line = "";
			continue;
			}
		}
	else if (smen_response == -1)			/* end of file detected */
		{
		Last_Line = "";
#ifdef	_PV_FILEIO_NOWRAP
		return (-1);						/* return an error */
#else
		fseek(smen_fd,0,0);					/* rewind to beginning */
#endif
		}
	/*  Always skip lines that start with a comment. */
	if ((strlen(Last_Line) != 0) &&
		(sindex(Last_Line,_PV_FILEIO_COMMENT) == 1))
		{
		Last_Line = "";
		continue;
		}
	/*  Strip comments off the end of the line. */
	if ((strlen(Last_Line) != 0) && 
		(sindex(Last_Line,_PV_FILEIO_COMMENT) != 0))
		{
		Last_Line = substr(Last_Line,1,sindex(Last_Line,_PV_FILEIO_COMMENT)-1);
		}
	/* Get rid of white space at the end of the line. */
	for (smen_index = strlen(Last_Line); smen_index > 0; smen_index--)
		{
		if (sindex(substr(Last_Line,smen_index,1),_PV_FILEIO_WHITESPACE) == 0)
			{
			break;		/* found non-white space */
			}
		}
	Last_Line = substr(Last_Line,1,smen_index);	/* truncate */

	/*  Recognize the special NULL line indicator.  */
	if (Last_Line == _PV_FILEIO_BLANKLINE)
		{
		Last_Line = "";
		break;
		}
	}
/*
	smen_next_field is set to 1 so that the NextField and NextSubField
	functions will work.
*/
smen_next_field = 1;
return (1);					/* return success */
}


/*
	The NextField function assumes that Last_Line contains the line
	of text to be parsed, and it returns the next sequential field.

	IndexedField returns the specified field by number, and initializes
	counters so that NextField will work properly.
*/
string func NextField()
{
	smen_next_subfield = 1;			/* set so that NextSubfield works */
#ifdef	_VU_VERSION			/* make sure we have version 4.0 or later */
#ifdef	_PV_FILEIO_ALLOWNULLS	/* insert NULL processing in needed */
	if (Allow_Nulls)
		{
		Last_Field = SubField(Last_Line,_PV_FILEIO_FIELD,smen_next_field++);
		if (smen_next_field > smen_sf_fields)
			{
			Final_Field = 1;
			}
		else
			{
			Final_Field = 0;
			}
		}
	else
		{
#endif
#endif
		
		Last_Field = subfield(Last_Line,_PV_FILEIO_FIELD,smen_next_field++);

#ifdef	_VU_VERSION			/* make sure we have version 4.0 or later */
#ifdef	_PV_FILEIO_ALLOWNULLS	/* insert NULL processing in needed */
		}
#endif
#endif

	return (Last_Field);
}

string func IndexedField(smen_if_index)
{
	smen_next_field = smen_if_index;
	return (NextField());
}

/*
	The NextSubField and IndexedSubField functions return the next subfield
	within the last field read.
*/
string func NextSubField()
{
#ifdef	_VU_VERSION				/* make sure we have version 4.0 or later */
#ifdef	_PV_FILEIO_ALLOWNULLS	/* null field processing if requested. */
	if (Allow_Nulls)
		{
		Last_SubField = SubField(Last_Field,_PV_FILEIO_SUBFIELD,
							 		smen_next_subfield++);
		if (smen_next_subfield > smen_sf_fields)
			{
			Final_SubField = 1;
			}
		else
			{
			Final_SubField = 0;
			}
		}
	else
		{
#endif
#endif

		Last_SubField = subfield(Last_Field,_PV_FILEIO_SUBFIELD,
							 		smen_next_subfield++);

#ifdef  _VU_VERSION             /* make sure we have version 4.0 or later */
#ifdef  _PV_FILEIO_ALLOWNULLS   /* null field processing if requested. */
		}
#endif
#endif
	return (Last_SubField);
}

string func IndexedSubField(smen_if_index)
{
	smen_next_subfield = smen_if_index;
	return (NextSubField());
}

/*
	The macro SHARED_READ implements locking and exclusive reading of a file
	by multiple users.  Each user will get the next line in the file when
	it implements the function.  The mechanism employed uses two shared
	variables.  One locks the file, while the other one keeps track of the
	next offset.  A "prefix" string identifies this pair (ie, if the prefix
	was "xxx", they would be xxx_lock, and xxx_offset).  The shared variable
	pair must be defined external to this include file.
*/
#define	SHARED_READ( FILE, PREFIX )	wait(&PREFIX##_lock,0,0,1);			\
									sme_sr_fd = open(FILE, "r");			\
									fseek(sme_sr_fd,PREFIX##_offset,0);	\
									ReadLine(sme_sr_fd);					\
									PREFIX##_offset = ftell(sme_sr_fd);	\
									close(sme_sr_fd);						\
									PREFIX##_lock--;

#endif					/* matches #ifndef _PV_FILEIO at top of file */
